use std::{collections::HashSet, hash::Hash, rc::Rc};

use proc_macro2::{Ident, Span};

use btrun::model::End;

use crate::{
    error::{self, MaybeErr},
    model::{ActorKind, MsgInfo, ProtocolModel, ValueKind},
    parsing::{DestinationState, GetSpan, LeftJoin, State},
};

impl ProtocolModel {
    #[allow(dead_code)]
    pub(crate) fn validate(&self) -> syn::Result<()> {
        self.all_states_declared_and_used()
            .combine(self.receivers_and_senders_matched())
            .combine(self.no_undeliverable_msgs())
            .combine(self.replies_expected())
            .combine(self.no_unobservable_states())
            .combine(self.no_ambiguous_trans())
            .combine(self.no_exit_races())
            .combine(self.no_invalid_out_state_provenance())
            .combine(self.no_invalid_out_msg_provenance())
            .into()
    }

    /// Verifies that every state which is declared is actually used.
    fn all_states_declared_and_used(&self) -> MaybeErr {
        let end = Ident::new(End::ident(), Span::call_site());
        let mut declared: HashSet<&Ident> = HashSet::new();
        declared.insert(&end);
        for actor_def in self.def().actor_defs.iter() {
            for state in actor_def.states.all() {
                declared.insert(state);
            }
        }
        let mut used: HashSet<&Ident> = HashSet::with_capacity(declared.len());
        for transition in self.def().transitions.iter() {
            let in_state = &transition.in_state;
            used.insert(&in_state.state_trait);
            used.extend(in_state.states_array_owned().map(|ident| ident.as_ref()));
            if let Some(in_msg) = transition.in_msg() {
                used.extend(in_msg.owned_states().map(|ident| ident.as_ref()));
            }
            for out_states in transition.out_states.as_ref().iter() {
                used.insert(&out_states.state_trait);
                used.extend(out_states.states_array_owned().map(|ident| ident.as_ref()));
            }
            // We don't have to check the states referred to in out_msgs because the
            // receivers_and_senders_matched method ensures that each of these exists in a receiver
            // position.
        }
        let undeclared: MaybeErr = used
            .difference(&declared)
            .map(|ident| syn::Error::new(ident.span(), error::msgs::UNDECLARED_STATE))
            .collect();
        let unused: MaybeErr = declared
            .difference(&used)
            .filter(|ident| **ident != End::ident())
            .map(|ident| syn::Error::new(ident.span(), error::msgs::UNUSED_STATE))
            .collect();
        undeclared.combine(unused)
    }

    /// Ensures that the recipient state for every sent message has a receiving transition
    /// defined, and every receiver has a sender. Note that each message isn't required to have a
    /// unique sender or a unique receiver, just that at least one of each much be defined.
    fn receivers_and_senders_matched<'s>(&'s self) -> MaybeErr {
        /// Represents a message sender or receiver.
        ///
        /// This type is essentially just a tuple of references, but was created so a [Hash]
        /// implementation could be defined.
        #[cfg_attr(test, derive(Debug))]
        struct MsgEndpoint<'a> {
            state: &'a State,
            msg_info: &'a MsgInfo,
        }

        impl<'a> MsgEndpoint<'a> {
            fn new(state: &'a State, msg_info: &'a MsgInfo) -> Self {
                Self { state, msg_info }
            }
        }

        impl<'a> PartialEq for MsgEndpoint<'a> {
            fn eq(&self, other: &Self) -> bool {
                self.state.state_trait == other.state.state_trait
                    && self.msg_info.def().msg_type == self.msg_info.def().msg_type
                    && self.msg_info.is_reply() == self.msg_info.is_reply()
            }
        }

        impl<'a> Eq for MsgEndpoint<'a> {}

        impl<'a> Hash for MsgEndpoint<'a> {
            fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
                self.state.state_trait.hash(state);
                self.msg_info.def().msg_type.hash(state);
                self.msg_info.is_reply().hash(state);
            }
        }

        #[cfg(test)]
        impl<'a> std::fmt::Display for MsgEndpoint<'a> {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                write!(
                    f,
                    "({}, {})",
                    self.state.state_trait,
                    self.msg_info.msg_name()
                )
            }
        }

        let msgs = self.msg_lookup();
        let mut outgoing: HashSet<MsgEndpoint<'s>> = HashSet::new();
        let mut incoming: HashSet<MsgEndpoint<'s>> = HashSet::new();
        for actor in self.actors_iter() {
            let methods = actor
                .states()
                .values()
                .flat_map(|state| state.methods().values());
            for method in methods {
                let transition = method.def();
                if let Some(msg) = transition.in_msg() {
                    let msg_info = msgs.lookup(msg);
                    incoming.insert(MsgEndpoint::new(&transition.in_state, msg_info));
                }
                for dest in transition.out_msgs.as_ref().iter() {
                    let dest_state = match &dest.state {
                        DestinationState::Individual(dest_state) => dest_state,
                        DestinationState::Service(dest_state) => dest_state,
                    };
                    let msg_info = self.msg_lookup().lookup(&dest.msg);
                    outgoing.insert(MsgEndpoint::new(dest_state, msg_info));
                    if actor.kind().is_client() {
                        if let Some(reply) = msg_info.reply() {
                            incoming.insert(MsgEndpoint::new(&transition.in_state, reply));
                        }
                    }
                }
            }
        }
        let extra_senders: MaybeErr = outgoing
            .difference(&incoming)
            .map(|endpoint| {
                syn::Error::new(
                    endpoint.msg_info.def().span(),
                    error::msgs::UNMATCHED_OUTGOING,
                )
            })
            .collect();
        let extra_receivers: MaybeErr = incoming
            .difference(&outgoing)
            .map(|endpoint| {
                syn::Error::new(
                    endpoint.msg_info.def().span(),
                    error::msgs::UNMATCHED_INCOMING,
                )
            })
            .collect();
        extra_senders.combine(extra_receivers)
    }

    /// Checks that messages are only sent to destinations which are either services, states
    /// which are owned by the sender, listed in the output states, or that the message is a
    /// reply.
    fn no_undeliverable_msgs(&self) -> MaybeErr {
        let mut err = MaybeErr::none();
        for transition in self.def().transitions.iter() {
            let mut allowed_states: Option<HashSet<&Ident>> = None;
            for dest in transition.out_msgs.as_ref().iter() {
                if dest.msg.is_reply() {
                    continue;
                }
                match &dest.state {
                    DestinationState::Service(_) => continue,
                    DestinationState::Individual(dest_state) => {
                        let allowed_from_msg = transition
                            .in_msg()
                            .into_iter()
                            .flat_map(|msg| msg.states_array.as_ref())
                            .flat_map(|array| array.all())
                            .map(|ptr| ptr.as_ref());
                        let allowed_from_state = transition
                            .in_state
                            .states_array_all()
                            .map(|ident| ident.as_ref())
                            .chain(allowed_from_msg);
                        let allowed = allowed_states.get_or_insert_with(|| {
                            transition
                                .out_states
                                .as_ref()
                                .iter()
                                .map(|state| state.state_trait.as_ref())
                                .chain(allowed_from_state)
                                .collect()
                        });
                        if !allowed.contains(dest_state.state_trait.as_ref()) {
                            err = err.combine(
                                syn::Error::new(
                                    dest_state.state_trait.span(),
                                    error::msgs::UNDELIVERABLE,
                                )
                                .into(),
                            );
                        }
                    }
                }
            }
        }
        err
    }

    /// Verifies that exactly one reply is sent in response to a previously sent message.
    fn replies_expected(&self) -> MaybeErr {
        let mut err = MaybeErr::none();
        for transition in self.def().transitions.iter() {
            let replies: Vec<_> = transition
                .out_msgs
                .as_ref()
                .iter()
                .map(|dest| &dest.msg)
                .filter(|msg| msg.is_reply())
                .collect();
            if replies.is_empty() {
                continue;
            }
            if replies.len() > 1 {
                err = err.combine(
                    replies
                        .iter()
                        .map(|reply| {
                            syn::Error::new(reply.msg_type.span(), error::msgs::MULTIPLE_REPLIES)
                        })
                        .collect(),
                );
            }
            if transition.in_msg().is_none() {
                err = err.combine(
                    replies
                        .iter()
                        .map(|reply| {
                            syn::Error::new(reply.msg_type.span(), error::msgs::INVALID_REPLY)
                        })
                        .collect(),
                );
            }
        }
        err
    }

    /// Checks that there are no client states which are only receiving replies. Such states can't
    /// be observed, because the methods which sent the original messages will return their replies.
    fn no_unobservable_states(&self) -> MaybeErr {
        self.actors_iter()
            .filter(|actor| actor.kind().is_client())
            .flat_map(|actor| actor.states().values())
            .filter(|state| {
                !state.methods().is_empty()
                    && state.methods().values().all(|method| {
                        if let Some(in_msg) = method.def().in_msg() {
                            in_msg.is_reply()
                        } else {
                            false
                        }
                    })
            })
            .map(|state| syn::Error::new(state.span(), error::msgs::UNOBSERVABLE_STATE))
            .collect()
    }

    /// Checks that there are no two transitions defined in a non-client state which do not receive
    /// messages. Such transition create ambiguity because we don't know which of the two
    /// transitions to execute before entering the non-client's loop.
    fn no_ambiguous_trans(&self) -> MaybeErr {
        self.actors_iter()
            .flat_map(|actor| {
                if matches!(actor.kind(), ActorKind::Client) {
                    return None;
                }
                let no_input_methods: Vec<_> = actor
                    .states()
                    .values()
                    .flat_map(|state| state.methods().values())
                    .filter(|method| method.msg_received_input().is_none())
                    .collect();
                if no_input_methods.len() > 1 {
                    Some(no_input_methods)
                } else {
                    None
                }
            })
            .map(|no_input_methods| {
                let mut iter = no_input_methods.into_iter();
                let mut span = iter.next().unwrap().span();
                for method in iter {
                    span = span.left_join(method.span());
                }
                syn::Error::new(span, error::msgs::AMBIGUOUS_TRANS)
            })
            .collect()
    }

    /// Ensures that no race conditions exist between the delivery of exit messages by the runtime
    /// and application messages.
    fn no_exit_races(&self) -> MaybeErr {
        self.methods_iter()
            .filter(|method| {
                let next_state_name =
                    if let ValueKind::State { def, .. } = method.next_state().kind() {
                        def.state_trait.as_ref()
                    } else {
                        panic!(
                            "Next state output of method '{}' is not actually a state.",
                            method.name()
                        );
                    };
                next_state_name == End::ident()
            })
            .flat_map(|method| {
                let owned: HashSet<_> = method
                    .def()
                    .owned_states()
                    .map(|ptr| ptr.as_ref())
                    .collect();
                method.def().out_msgs.as_ref().iter().flat_map(move |dest| {
                    if let DestinationState::Individual(state) = &dest.state {
                        let dest_state = state.state_trait.as_ref();
                        if owned.contains(dest_state) {
                            Some(dest)
                        } else {
                            None
                        }
                    } else {
                        None
                    }
                })
            })
            .map(|dest| syn::Error::new(dest.span(), error::msgs::EXIT_RACE))
            .collect()
    }

    /// Checks that ownership granted to output states given when the input
    /// state or the input message in the same transition owns the granted state.
    fn no_invalid_out_state_provenance(&self) -> MaybeErr {
        self.actors_iter()
            .flat_map(|actor| actor.states().values().map(move |state| (actor, state)))
            .flat_map(|(actor, state)| state.methods().values().map(move |method| (actor, method)))
            .flat_map(|(actor, method)| {
                let valid_owned_states: HashSet<&Rc<Ident>> = if actor.kind().is_client() {
                    // A client is allowed to get ownership from the replies to the messages it's
                    // sending.
                    let states_from_reply = method
                        .def()
                        .out_msgs
                        .as_ref()
                        .iter()
                        .filter(|out_msg| {
                            let msg_info = self.msg_lookup().lookup(&out_msg.msg);
                            msg_info.is_call()
                        })
                        .flat_map(|out_msg| {
                            let receiver_state_name =
                                out_msg.state.state_ref().state_trait.as_ref();
                            let receiver_state = self.get_state(receiver_state_name);
                            let msg = &out_msg.msg;
                            let receiver_method = receiver_state
                                .methods()
                                .values()
                                .find(|method| method.def().in_msg() == Some(msg))
                                .unwrap_or_else(|| {
                                    panic!(
                                    "There is no receiver for message '{}' sent by method '{}'.",
                                    msg.msg_type.as_ref(),
                                    method.name(),
                                )
                                });
                            let reply_msg = receiver_method
                                .def()
                                .out_msgs
                                .as_ref()
                                .iter()
                                .find(|reply_msg| {
                                    reply_msg.msg.is_reply()
                                        && reply_msg.msg.msg_type.as_ref() == msg.msg_type.as_ref()
                                })
                                .unwrap_or_else(|| {
                                    panic!(
                                        "Method '{}' is not sending reply to message '{}'.",
                                        receiver_method.name(),
                                        msg.msg_type.as_ref(),
                                    )
                                });
                            reply_msg.msg.owned_states()
                        });
                    method
                        .def()
                        .owned_states()
                        .chain(states_from_reply)
                        .collect()
                } else {
                    method.def().owned_states().collect()
                };
                let valid_owned_states = Rc::new(valid_owned_states);
                method
                    .def()
                    .out_states
                    .as_ref()
                    .iter()
                    .flat_map(move |out_state| {
                        let valid_owned_states = valid_owned_states.clone();
                        out_state
                            .states_array_owned()
                            .filter(move |out_state_owned| {
                                !valid_owned_states.contains(out_state_owned)
                            })
                    })
            })
            .map(|state| syn::Error::new(state.span(), error::msgs::INVALID_OUT_STATE_PROVENANCE))
            .collect()
    }

    /// Checks that ownership granted to output messages is only given for states owned by the
    /// input state, input message, or which are output states in the same transition.
    fn no_invalid_out_msg_provenance(&self) -> MaybeErr {
        self.methods_iter()
            .flat_map(|method| {
                let out_msg_not_valid = |state: &&Rc<Ident>| {
                    method
                        .def()
                        .owned_states()
                        .chain(
                            method
                                .def()
                                .out_states
                                .as_ref()
                                .iter()
                                .map(|out_state| &out_state.state_trait),
                        )
                        .all(|owned_state| owned_state.as_ref() != state.as_ref())
                };
                method
                    .def()
                    .out_msgs
                    .as_ref()
                    .iter()
                    .flat_map(move |out_msg| out_msg.msg.owned_states().filter(out_msg_not_valid))
            })
            .map(|state| syn::Error::new(state.span(), error::msgs::INVALID_OUT_MSG_PROVENANCE))
            .collect()
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{
        error::{assert_err, assert_ok},
        parsing::{ActorDef, Dest, Message, NameDef, Protocol, Transition},
    };

    #[test]
    fn all_states_declared_and_used_ok() {
        let input = ProtocolModel::new(Protocol::minimal()).unwrap();

        let result = input.validate();

        assert_ok(result);
    }

    #[test]
    fn all_states_declared_and_used_end_not_used_ok() {
        const STATE_NAME: &str = "Init";
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Test"),
            [ActorDef::new("actor", [STATE_NAME])],
            [Transition::new(
                State::new(STATE_NAME, []),
                Some(Message::new("Activate", false, [])),
                [State::new(STATE_NAME, [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_ok(result);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("Next", [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_err(result, error::msgs::UNDECLARED_STATE);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_out_state_owned_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init", "Next"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("Init", []), State::new("Next", ["Undeclared"])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_err(result, error::msgs::UNDECLARED_STATE);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_in_state_owned_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init", "Next"])],
            [Transition::new(
                State::new("Init", ["Undeclared"]),
                Some(Message::new("Activate", false, [])),
                [State::new("Next", [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_err(result, error::msgs::UNDECLARED_STATE);
    }

    #[test]
    fn all_states_declared_and_used_unused_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Unused"),
            [ActorDef::new("actor", ["Init", "Extra"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("End", [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_err(result, error::msgs::UNUSED_STATE);
    }

    #[test]
    fn receivers_and_senders_matched_ok() {
        let input = ProtocolModel::new(Protocol::minimal()).unwrap();

        let result = input.receivers_and_senders_matched();

        assert_ok(result);
    }

    #[test]
    fn receivers_and_senders_call_msg_reused_for_non_call_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("OwnedTypes"),
            [
                ActorDef::new("server", ["Listening"]),
                ActorDef::new("client", ["Client"]),
                ActorDef::new("file", ["FileInit", "Opened"]),
                ActorDef::new("file_handle", ["FileHandle"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [
                        State::new("Client", []),
                        State::new("FileHandle", ["Opened"]),
                    ],
                    [Dest::new(
                        DestinationState::Service(State::new("Listening", [])),
                        Message::new("Open", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Listening", []),
                    Some(Message::new("Open", false, [])),
                    [State::new("Listening", []), State::new("FileInit", [])],
                    [
                        Dest::new(
                            DestinationState::Individual(State::new("Client", [])),
                            Message::new("Open", true, ["Opened"]),
                        ),
                        // Note that the same "Open" message is being used here, but that the
                        // FileInit state does not send a reply. This should be allowed.
                        Dest::new(
                            DestinationState::Individual(State::new("FileInit", [])),
                            Message::new("Open", false, []),
                        ),
                    ],
                ),
                Transition::new(
                    State::new("FileInit", []),
                    Some(Message::new("Open", false, [])),
                    [State::new("Opened", [])],
                    [],
                ),
            ],
        ))
        .unwrap();

        let result = input.receivers_and_senders_matched();

        assert_ok(result);
    }

    #[test]
    fn receivers_and_senders_matched_unmatched_sender_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Unbalanced"),
            [
                ActorDef::new("client", ["Init"]),
                ActorDef::new("server", ["Server"]),
            ],
            [
                Transition::new(
                    State::new("Init", []),
                    None,
                    [State::new("Init", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("UnexpectedMsg", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Init", []),
                    None,
                    [State::new("Init", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("ExpectedMsg", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("ExpectedMsg", false, [])),
                    [State::new("End", [])],
                    [],
                ),
            ],
        ))
        .unwrap();

        let result = input.receivers_and_senders_matched();

        assert_err(result, error::msgs::UNMATCHED_OUTGOING);
    }

    #[test]
    fn receivers_and_senders_matched_unmatched_receiver_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Unbalanced"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("NotExists", false, [])),
                [State::new("Init", [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.receivers_and_senders_matched();

        assert_err(result, error::msgs::UNMATCHED_INCOMING);
    }

    #[test]
    fn receivers_and_senders_matched_servers_must_explicitly_receive_replies_err() {
        // Only client actors are allowed to implicitly receive replies.
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Conversation"),
            [
                ActorDef::new("alice", ["Alice"]),
                ActorDef::new("bob", ["Bob"]),
            ],
            [
                Transition::new(
                    State::new("Alice", []),
                    None,
                    [State::new("Alice", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Bob", [])),
                        Message::new("Greeting", false, []),
                    )],
                ),
                // Notice that because Bob only has transitions which handle messages, bob is a
                // server actor.
                Transition::new(
                    State::new("Bob", []),
                    Some(Message::new("Greeting", false, [])),
                    [State::new("Bob", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Alice", [])),
                        Message::new("Query", false, []),
                    )],
                ),
                // Alice is sending a Query::Reply to Bob, but because he does not have a
                // transition which accepts that message type, this will be an unmatched outgoing
                // error.
                Transition::new(
                    State::new("Alice", []),
                    Some(Message::new("Query", false, [])),
                    [State::new("End", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Bob", [])),
                        Message::new("Query", true, []),
                    )],
                ),
            ],
        ))
        .unwrap();

        let result = input.receivers_and_senders_matched();

        assert_err(result, error::msgs::UNMATCHED_OUTGOING);
    }

    #[test]
    fn no_undeliverable_msgs_ok() {
        let input = ProtocolModel::new(Protocol::minimal()).unwrap();

        let result = input.no_undeliverable_msgs();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_reply_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["Listening", "Client"])],
            [Transition::new(
                State::new("Listening", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Listening", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Client", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.no_undeliverable_msgs();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_service_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeliverable"),
            [
                ActorDef::new("client", ["Client"]),
                ActorDef::new("server", ["Server"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Client", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("Msg", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("Msg", false, [])),
                    [State::new("End", [])],
                    [],
                ),
            ],
        ))
        .unwrap();

        let result = input.no_undeliverable_msgs();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_owned_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["FileClient", "FileHandle"])],
            [Transition::new(
                State::new("FileClient", ["FileHandle"]),
                None,
                [State::new("FileClient", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("FileHandle", [])),
                    Message::new("FileOp", false, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.no_undeliverable_msgs();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_sending_when_handling_owned_msg_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("EndAndSend"),
            [
                ActorDef::new("client", ["Client", "Subbed"]),
                ActorDef::new("server", ["Server"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Subbed", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("GiveSelf", false, ["Subbed"]),
                    )],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("GiveSelf", false, ["Subbed"])),
                    [State::new("Server", ["Subbed"])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Subbed", [])),
                        Message::new("FinalOrders", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Subbed", []),
                    Some(Message::new("FinalOrders", false, [])),
                    [State::new("End", [])],
                    [],
                ),
            ],
        ))
        .unwrap();

        let result = input.validate();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Client", []),
                None,
                [State::new("Client", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Server", [])),
                    Message::new("Msg", false, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.no_undeliverable_msgs();

        assert_err(result, error::msgs::UNDELIVERABLE);
    }

    #[test]
    fn replies_expected_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Server", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Server", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Client", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.replies_expected();

        assert_ok(result);
    }

    #[test]
    fn replies_expected_invalid_reply_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Client", []),
                None,
                [State::new("Client", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Server", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.replies_expected();

        assert_err(result, error::msgs::INVALID_REPLY);
    }

    #[test]
    fn replies_expected_multiple_replies_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "OtherClient", "Server"])],
            [Transition::new(
                State::new("Server", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Server", [])],
                [
                    Dest::new(
                        DestinationState::Individual(State::new("Client", [])),
                        Message::new("Msg", true, []),
                    ),
                    Dest::new(
                        DestinationState::Individual(State::new("OtherClient", [])),
                        Message::new("Msg", true, []),
                    ),
                ],
            )],
        ))
        .unwrap();

        let result = input.replies_expected();

        assert_err(result, error::msgs::MULTIPLE_REPLIES);
    }

    #[test]
    fn no_unobservable_states_reply_to_client_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Test"),
            [
                ActorDef::new("server", ["Server"]),
                ActorDef::new("client", ["Client"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Client", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("Msg", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("Msg", false, [])),
                    [State::new("Server", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Client", [])),
                        Message::new("Msg", true, []),
                    )],
                ),
            ],
        ))
        .unwrap();

        let result = input.validate();

        assert_ok(result);
    }

    /// Only one method in a non-client actor state is allowed to have no input message, because
    /// otherwise there would be ambiguity as to which method to call before the actor entered
    /// its loop.
    #[test]
    fn no_ambiguous_trans_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("MultipleNoInputTransitions"),
            [
                ActorDef::new("client", ["Client"]),
                ActorDef::new("server", ["Server"]),
                ActorDef::new("worker", ["Worker"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Client", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("Msg", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("Msg", false, [])),
                    [State::new("Server", []), State::new("Worker", [])],
                    [],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("Halt", false, [])),
                    [State::new("End", [])],
                    [],
                ),
                // Notice that worker sends two different messages from its initial state. This
                // crates a ambiguity because we don't know which one should be called in the
                // generated code.
                Transition::new(
                    State::new("Worker", []),
                    None,
                    [State::new("Worker", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("Msg", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Worker", []),
                    None,
                    [State::new("End", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("Halt", false, []),
                    )],
                ),
            ],
        ))
        .unwrap();

        let result = input.no_ambiguous_trans();

        assert_err(result, error::msgs::AMBIGUOUS_TRANS);
    }

    #[test]
    fn no_exit_races_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("EndAndSend"),
            [
                ActorDef::new("client", ["Client", "Subbed"]),
                ActorDef::new("server", ["Server"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Subbed", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("GiveSelf", false, ["Subbed"]),
                    )],
                ),
                // Notice that Server sends FinalOrders to Subbed and immediately exits. This
                // creates a race condition between the owner exited message sent by the runtime
                // and the FinalOrders message.
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("GiveSelf", false, ["Subbed"])),
                    [State::new("End", ["Subbed"])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Subbed", [])),
                        Message::new("FinalOrders", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Subbed", []),
                    Some(Message::new("FinalOrders", false, [])),
                    [State::new("End", [])],
                    [],
                ),
            ],
        ))
        .unwrap();

        let result = input.validate();

        assert_err(result, error::msgs::EXIT_RACE);
    }

    #[test]
    fn no_invalid_out_state_provenance_ok() {
        let input = ProtocolModel::new(Protocol::minimal()).unwrap();

        let result = input.no_invalid_out_state_provenance();

        assert_ok(result);
    }

    #[test]
    fn no_invalid_out_state_provenance_ownership_from_reply_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("OwnershipFromReply"),
            [
                ActorDef::new("client", ["Client"]),
                ActorDef::new("handle", ["Handle"]),
                ActorDef::new("server", ["Server"]),
                ActorDef::new("worker", ["Worker"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Client", []), State::new("Handle", ["Worker"])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("Open", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("Open", false, [])),
                    [State::new("Server", []), State::new("Worker", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Client", [])),
                        Message::new("Open", true, ["Worker"]),
                    )],
                ),
            ],
        ))
        .unwrap();

        let result = input.validate();

        assert_ok(result);
    }

    #[test]
    fn no_invalid_out_state_provenance_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Test"),
            [
                ActorDef::new("server", ["Server"]),
                ActorDef::new("client", ["Client"]),
                ActorDef::new("worker", ["Worker"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Client", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("Msg", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("Msg", false, [])),
                    [
                        State::new("Server", []),
                        // Note that this transition doesn't have the right to give ownership of
                        // Client.
                        State::new("Worker", ["Client"]),
                    ],
                    [],
                ),
            ],
        ))
        .unwrap();

        let result = input.validate();

        assert_err(result, error::msgs::INVALID_OUT_STATE_PROVENANCE);
    }

    #[test]
    fn no_invalid_out_msgs_provenance_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Test"),
            [
                ActorDef::new("server", ["Server"]),
                ActorDef::new("client", ["Client"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Client", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Server", [])),
                        Message::new("Msg", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Server", []),
                    Some(Message::new("Msg", false, [])),
                    [State::new("Server", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Client", [])),
                        // Note that the this transition has no right to give ownership of Client.
                        Message::new("Msg", true, ["Client"]),
                    )],
                ),
            ],
        ))
        .unwrap();

        let result = input.validate();

        assert_err(result, error::msgs::INVALID_OUT_MSG_PROVENANCE);
    }
}
